//rww - model decals

#include "quakedef.h"

typedef struct clDecalSet_s
{
	modelDecal_t	*decals;
} clDecalSet_t;

clDecalSet_t g_clDecals[MAX_CLIENTS];

//free attached decals
void Decals_Free(clDecalSet_t *obj)
{
	modelDecal_t *d = obj->decals;
	while (d)
	{
		modelDecal_t *nd = d->next;
		free(d);
		d = nd;
	}
	obj->decals = NULL;
}

//free all decals on client sets
void Decals_FreeAll(void)
{
	int i;
	for (i = 0; i < MAX_CLIENTS; i++)
	{
		Decals_Free(&g_clDecals[i]);
	}
}

//insert a decal
void Decals_Insert(clDecalSet_t *obj, modelDecal_t *dec)
{
	dec->next = obj->decals;
	dec->last = NULL;
	if (obj->decals)
	{
		obj->decals->last = dec;
	}
	obj->decals = dec;
}

//remove a decal
void Decals_Remove(clDecalSet_t *obj, modelDecal_t *dec)
{
	if (dec->last)
	{
		dec->last->next = dec->next;
	}
	else
	{ //if i am the root, assign the root to my next
		obj->decals = dec->next;
	}
	if (dec->next)
	{
		dec->next->last = dec->last;
	}
	free(dec);
}

//plane distance
static __inline float Decals_PlaneDist(const float *pl, const float *p)
{
	return pl[0]*p[0] + pl[1]*p[1] + pl[2]*p[2] + pl[3];
}

//project point onto plane
static __inline void Decals_ProjectOntoPlane(const float *plane, const float *pos, float *out)
{
	float d = Decals_PlaneDist(plane, pos);
	out[0] = pos[0] - plane[0]*d;
	out[1] = pos[1] - plane[1]*d;
	out[2] = pos[2] - plane[2]*d;
}

//create a plane from a position and a normal vector
static __inline void Decals_PlaneFromPosAndNormal(const float *pos, const float *normal, float *plane, qboolean invert)
{
	if (invert)
	{
		plane[0] = -normal[0];
		plane[1] = -normal[1];
		plane[2] = -normal[2];
	}
	else
	{
		plane[0] = normal[0];
		plane[1] = normal[1];
		plane[2] = normal[2];
	}
	VectorNormalize(plane);
	plane[3] = -(pos[0]*plane[0] + pos[1]*plane[1] + pos[2]*plane[2]);
}

//construct a plane from points
static __inline void Decals_PlaneFromPoints(float *_p0, float *_p1, float *_p2,
					 float *plane)
{
	float v0[3], v1[3];
	float n[3];
	VectorSubtract(_p0, _p1, v0);
	VectorSubtract(_p2, _p1, v1);
	n[0] = v1[1] * v0[2] - v1[2] * v0[1];
	n[1] = v1[2] * v0[0] - v1[0] * v0[2];
	n[2] = v1[0] * v0[1] - v1[1] * v0[0];
	VectorNormalize(n);
	plane[0] = n[0];
	plane[1] = n[1];
	plane[2] = n[2];
	plane[3] = -(_p0[0] * plane[0] + _p0[1] * plane[1] + _p0[2] * plane[2]);
}

//check if the sphere is on the line
static __inline qboolean Decals_DirectSphereLine(float *sphereCenter, float sphereRad, float *p1, float *p2)
{
	float d[3];
	float lineVec[3];
	float lineLen;
	float sphereDist;
	float sphereOnLine[3];
	VectorSubtract(p2, p1, lineVec);
	//lineLen = Length(lineVec);
	lineLen = VectorNormalize(lineVec);

	VectorSubtract(sphereCenter, p1, d);
	sphereDist = Length(d);
	if (sphereDist > (lineLen+sphereRad))
	{ //goes off the other end of the line
		return false;
	}
	//VectorNormalize(d);

	VectorMA(p1, sphereDist, lineVec, sphereOnLine);

	VectorSubtract(sphereOnLine, sphereCenter, d);
	//if (Length(d) <= sphereRad)
	if ((d[0]*d[0] + d[1]*d[1] + d[2]*d[2]) <= (sphereRad*sphereRad))
	{
		return true;
	}

	return false;
}

//see if a point is inside a given triangle
static __inline qboolean Decals_CheckPointInTri(float *v1, float *v2, float *v3, float *point)
{
	float u[3], v[3], w[3];
	float uU, uV, vV, wU, wV;
	float d, invD;
	float s, t;

	VectorSubtract(v2, v1, u);
	VectorSubtract(v3, v1, v);
	VectorSubtract(point, v1, w);
	uU = DotProduct(u, u);
	uV = DotProduct(u, v);
	vV = DotProduct(v, v);
	wU = DotProduct(w, u);
	wV = DotProduct(w, v);
	d = uV*uV - uU*vV;
	if (d == 0.0f)
	{
		invD = 0.0f;
	}
	else
	{
		invD = 1.0f/d;
	}
	s = (uV*wV - vV*wU)*invD;
	if (s < 0.0f || s > 1.0f)
	{
		return false;
	}
	t = (uV*wU - uU*wV)*invD;
	if (t < 0.0f || (s+t) > 1.0f)
	{
		return false;
	}

	return true;
}

//given the verts and plane of a triangle, see if a sphere is overlapping them
qboolean Decals_SphereOverlapsTri(float *sphereCenter, float sphereRad, float *triVerts[3], float *plane)
{
	int i;
	//check if the sphere is inside the tri plane
	float s = plane[0]*sphereCenter[0] + plane[1]*sphereCenter[1] + plane[2]*sphereCenter[2] + plane[3];
	if (s <= sphereRad && s >= -sphereRad)
	{
		if (Decals_CheckPointInTri(triVerts[0], triVerts[1], triVerts[2], sphereCenter))
		{
			return true;
		}
	}

	//check if a vert is within the radius
	for (i = 0; i < 3; i++)
	{
		float d[3];
		VectorSubtract(triVerts[i], sphereCenter, d);
		if (Length(d) <= sphereRad)
		{
			return true;
		}
	}

	//check if any of the edges intersect
	for (i = 0; i < 3; i++)
	{
		if (Decals_DirectSphereLine(sphereCenter, sphereRad, triVerts[i], triVerts[(i+1)%3]))
		{
			return true;
		}
	}

	return false;
}

//decal projection
void Decals_ProjectDecal(entity_t *modelSceneObj, clDecalSet_t *clDec, float *start, float *end, float decalSize,
							  int decalTex, double lifeTime, double fadeTime)
{
	float localStart[3], localEnd[3], localDir[3];
	float collPlane[4];
	float traceDist;
	float planeRight[3], planeUp[3];
	float t[3];
	#define DECAL_IDX_BUFFER_SIZE 4096
	int idxBuffer[DECAL_IDX_BUFFER_SIZE*3]; //projecting onto more than 4096 tris is a little crazy anyway
	aliashdr_t *hdr;
	entity_t *oldCur;
	int frame = modelSceneObj->frame;
	int pose, numposes;
	float interval;

	if (!modelSceneObj->model || modelSceneObj->model->type != mod_alias)
	{
		return;
	}

	hdr = (aliashdr_t *)Mod_Extradata(modelSceneObj->model);
	if ((frame >= hdr->numframes) || (frame < 0))
	{
		frame = 0;
	}
	pose = hdr->frames[frame].firstpose;
	numposes = hdr->frames[frame].numposes;
	if (numposes > 1)
	{
		interval = hdr->frames[frame].interval;
		pose += (int)(cl.time / interval) % numposes;
	}

	//make sure the ent's inverse matrix is up to date
	oldCur = currententity;
	currententity = modelSceneObj;
	GL_SetupCurEntityForDLight(NULL);
	currententity = oldCur;

	//transform start/end to model space
	TransformPointByMatrix4x4(modelSceneObj->invModelMat, start, localStart);
	TransformPointByMatrix4x4(modelSceneObj->invModelMat, end, localEnd);
	VectorSubtract(localEnd, localStart, localDir);
	if (localDir[0] == 0.0f && localDir[1] == 0.0f && localDir[2] == 0.0f)
	{
		localDir[2] = 1.0f;
	}

	traceDist = Length(localDir);
	VectorNormalize(localDir);

	//create a collision plane to base decal projection upon
	Decals_PlaneFromPosAndNormal(localStart, localDir, collPlane, true);

	vectoangles(collPlane, t);
	AngleVectors(t, 0, planeRight, planeUp);

	if (1)
	{
		modelVert_t *verts = (modelVert_t *)(hdr->batchVerts+(3*hdr->numverts*pose));
		modelTriFace_t *faces = (modelTriFace_t *)hdr->tris;
		int numFaces = hdr->numtris;
		int idxNum;
		int i;
		if (!verts || !faces || numFaces <= 0)
		{ //no verts/tris to work with
			return;
		}

		idxNum = 0;
		for (i = 0; i < numFaces; i++)
		{
			modelTriFace_t *face = faces+i;
			float *triVerts[3] = {
				(float *)(verts+face->a),
				(float *)(verts+face->b),
				(float *)(verts+face->c)
			};
			float planeEq[4];
			const float colFudge = 120.0f;	//this is busted due to laziness - need to do a proper
											//ray trace first and get the triangle impact point
			float triDirDot;

			Decals_PlaneFromPoints(triVerts[2], triVerts[1], triVerts[0], planeEq);

			triDirDot = DotProduct(localDir, planeEq);
			if (triDirDot > -0.1f)//-0.001f)
			{ //plane is not facing the projection direction
				continue;
			}

			if (Decals_SphereOverlapsTri(localStart, decalSize+colFudge, triVerts, planeEq))
			{ //this tri should then be included in the decal
				idxBuffer[(idxNum*3)+0] = face->a;
				idxBuffer[(idxNum*3)+1] = face->b;
				idxBuffer[(idxNum*3)+2] = face->c;
				idxNum++;
				if (idxNum >= DECAL_IDX_BUFFER_SIZE)
				{
					break;
				}
			}
		}
		if (idxNum > 0)
		{ //we got some, create a decal for the mesh object
			int idxSize = (idxNum*3)*sizeof(int);
			int tcSize = (hdr->numverts*2)*sizeof(float);
			modelDecal_t *dec;
			unsigned char *rawDec;
			int *decFaces;
			float *decTC;

			dec = (modelDecal_t *)malloc(sizeof(modelDecal_t)+idxSize+tcSize);
			memset(dec, 0, sizeof(modelDecal_t));
			rawDec = (unsigned char *)dec;

			dec->offIdx = sizeof(modelDecal_t);
			memcpy(rawDec+dec->offIdx, idxBuffer, idxSize);
			dec->offTC = dec->offIdx+idxSize;

			dec->numIdx = idxNum;
			dec->decTex = decalTex;
			dec->decalSize = decalSize;
			dec->pos[0] = localStart[0];
			dec->pos[1] = localStart[1];
			dec->pos[2] = localStart[2];
			dec->normal[0] = localDir[0];
			dec->normal[1] = localDir[1];
			dec->normal[2] = localDir[2];
			dec->createTime = realtime;
			dec->lifeSpan = lifeTime;
			dec->fadeTime = fadeTime;
			dec->modelOwner = modelSceneObj->model;

			//go through and figure out the texture coordinates based on the
			//collision plane
			decFaces = (int *)(rawDec+dec->offIdx);
			decTC = (float *)(rawDec+dec->offTC);
			for (i = 0; i < dec->numIdx; i++)
			{
				int j;
				int *face = decFaces+(i*3);
				float *triVerts[3] = {
					(float *)(verts+face[0]),
					(float *)(verts+face[1]),
					(float *)(verts+face[2])
				};
				float *triTC[3] = {
					(decTC+(face[0]*2)),
					(decTC+(face[1]*2)),
					(decTC+(face[2]*2))
				};

				for (j = 0; j < 3; j++)
				{
					float *projPlane = collPlane;
					float *projRight = planeRight;
					float *projUp = planeUp;
					float vertOnCol[3];
					float vertDif[3], vertDifX[3], vertDifY[3];
					float difX;
					float difY;

					//get the vert projected onto the collision plane
					Decals_ProjectOntoPlane(projPlane, triVerts[j], vertOnCol);

					VectorSubtract(vertOnCol, localStart, vertDif);
					
					//get the difference on the right and up axises to translate directly to s-t coordinates
					vertDifX[0] = vertDif[0]*projRight[0];
					vertDifX[1] = vertDif[1]*projRight[1];
					vertDifX[2] = vertDif[2]*projRight[2];
					vertDifY[0] = vertDif[0]*projUp[0];
					vertDifY[1] = vertDif[1]*projUp[1];
					vertDifY[2] = vertDif[2]*projUp[2];

					difX = vertDifX[0]+vertDifX[1]+vertDifX[2];
					difY = vertDifY[0]+vertDifY[1]+vertDifY[2];
					triTC[j][0] = difX/decalSize + 0.5f;
					triTC[j][1] = difY/decalSize + 0.5f;
				}
			}

			Decals_Insert(clDec, dec);
		}
	}
}

typedef struct defProj_s
{
	float			pos[3];
	float			dir[3];
	float			decalSize;
	int				decalTex;
	double			lifeTime;
	double			fadeTime;
	int				entIndex;
} defProj_t;

#define MAX_DEFERRED_PROJ	32
defProj_t g_defProj[MAX_DEFERRED_PROJ];
int g_numDefProj = 0;

//defer a projection for later
void Decal_DeferredProjection(float *pos, float *dir, float decalSize, int decalTex,
							  double lifeTime, double fadeTime, int entIndex)
{
	defProj_t *d;

	if (!decalTex || g_numDefProj >= MAX_DEFERRED_PROJ)
	{
		return;
	}

	d = &g_defProj[g_numDefProj++];
	VectorCopy(pos, d->pos);
	VectorCopy(dir, d->dir);
	d->decalSize = decalSize;
	d->decalTex = decalTex;
	d->lifeTime = lifeTime;
	d->fadeTime = fadeTime;
	d->entIndex = entIndex;
}

//check deferred projections
void Decal_CheckDeferred(void)
{
	int i;
	if (g_numDefProj <= 0)
	{
		return;
	}
	for (i = 0; i < cl_numvisedicts; i++)
	{
		entity_t *e = &cl_visedicts[i];
		int j;

		for (j = 0; j < g_numDefProj; j++)
		{
			defProj_t *d = &g_defProj[j];

			if (d->entIndex == e->gameentnum)
			{
				float start[3], end[3];
				float rayLen = 1.0f;
				VectorCopy(d->pos, start);
				start[0] += e->origin[0];
				start[1] += e->origin[1];
				start[2] += e->origin[2];
				VectorCopy(start, end);
				end[0] += d->dir[0]*rayLen;
				end[1] += d->dir[1]*rayLen;
				end[2] += d->dir[2]*rayLen;
				Decals_ProjectDecal(e, &g_clDecals[d->entIndex-1], start, end, d->decalSize, d->decalTex, d->lifeTime, d->fadeTime);
			}
		}
	}

	g_numDefProj = 0;
}

//render a decal on a model
void Decal_RenderDecal(entity_t *e, modelDecal_t *dec)
{
	int frame = e->frame;
	int pose, numposes;
	float interval;
	aliashdr_t *hdr;
	unsigned char *rawDec = (unsigned char *)dec;
	float decalFade = 1.0f;
	if (dec->lifeSpan != 0 && dec->fadeTime != 0)
	{ //check for a time-based fadeout
		double lifeDif = dec->lifeSpan-dec->fadeTime;
		double lifePoint = realtime-dec->createTime;
		if (lifePoint > lifeDif)
		{
			lifePoint -= lifeDif;
			decalFade = 1.0f-(lifePoint/dec->fadeTime);
		}
	}
	if (decalFade < 0.0f)
	{
		decalFade = 0.0f;
	}

	hdr = (aliashdr_t *)Mod_Extradata(e->model);
	if ((frame >= hdr->numframes) || (frame < 0))
	{
		frame = 0;
	}
	pose = hdr->frames[frame].firstpose;
	numposes = hdr->frames[frame].numposes;
	if (numposes > 1)
	{
		interval = hdr->frames[frame].interval;
		pose += (int)(cl.time / interval) % numposes;
	}

    glPushMatrix ();
	R_RotateForEntity (e);
	glTranslatef (hdr->scale_origin[0], hdr->scale_origin[1], hdr->scale_origin[2]);
	glScalef (hdr->scale[0], hdr->scale[1], hdr->scale[2]);

	GL_Bind(dec->decTex);
	glEnable(GL_BLEND);
	glDisable(GL_ALPHA_TEST);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glBlendFunc(GL_DST_COLOR, GL_ONE_MINUS_SRC_ALPHA);
	glColor4f(decalFade, decalFade, decalFade, decalFade);

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glVertexPointer(3, GL_FLOAT, 3 * sizeof(GL_FLOAT), (hdr->batchVerts+(3*hdr->numverts*pose)));

	glTexCoordPointer(2, GL_FLOAT, 2 * sizeof(GL_FLOAT), rawDec+dec->offTC);
	glDrawElements(GL_TRIANGLES, dec->numIdx*3, GL_UNSIGNED_INT, rawDec+dec->offIdx);

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);

	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glDisable(GL_BLEND);

	glPopMatrix();
}

//render client decals
void Decal_RenderClientDecals(void)
{
	int i;
	for (i = 0; i < cl_numvisedicts; i++)
	{
		entity_t *e = &cl_visedicts[i];
		if (e->gameentnum > 0 && e->gameentnum <= MAX_CLIENTS)
		{
			clDecalSet_t *d = &g_clDecals[e->gameentnum-1];
			modelDecal_t *dec = d->decals;
			while (dec)
			{
				modelDecal_t *next = dec->next;
				if (dec->modelOwner == e->model)
				{
					Decal_RenderDecal(e, dec);
				}
				else
				{
					Decals_Remove(d, dec);
				}
				dec = next;
			}
		}
	}

	//check for dying decals
	for (i = 0; i < MAX_CLIENTS; i++)
	{
		clDecalSet_t *d = &g_clDecals[i];
		modelDecal_t *dec = d->decals;
		while (dec)
		{
			modelDecal_t *next = dec->next;
			if (dec->lifeSpan != 0 &&
				realtime > (dec->createTime+dec->lifeSpan))
			{ //time to die
				Decals_Remove(d, dec);
			}
			dec = next;
		}
	}
}
