/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// mathlib.c -- math primitives

#include <math.h>
#include "quakedef.h"

void Sys_Error (char *error, ...);

vec3_t vec3_origin = {0,0,0};
int nanmask = 255<<23;

/*-----------------------------------------------------------------*/

#define DEG2RAD( a ) ( a * M_PI ) / 180.0F

void ProjectPointOnPlane( vec3_t dst, const vec3_t p, const vec3_t normal )
{
	float d;
	vec3_t n;
	float inv_denom;

	inv_denom = 1.0F / DotProduct( normal, normal );

	d = DotProduct( normal, p ) * inv_denom;

	n[0] = normal[0] * inv_denom;
	n[1] = normal[1] * inv_denom;
	n[2] = normal[2] * inv_denom;

	dst[0] = p[0] - d * n[0];
	dst[1] = p[1] - d * n[1];
	dst[2] = p[2] - d * n[2];
}

/*
** assumes "src" is normalized
*/
void PerpendicularVector( vec3_t dst, const vec3_t src )
{
	int	pos;
	int i;
	float minelem = 1.0F;
	vec3_t tempvec;

	/*
	** find the smallest magnitude axially aligned vector
	*/
	for ( pos = 0, i = 0; i < 3; i++ )
	{
		if ( fabs( src[i] ) < minelem )
		{
			pos = i;
			minelem = fabs( src[i] );
		}
	}
	tempvec[0] = tempvec[1] = tempvec[2] = 0.0F;
	tempvec[pos] = 1.0F;

	/*
	** project the point onto the plane defined by src
	*/
	ProjectPointOnPlane( dst, tempvec, src );

	/*
	** normalize the result
	*/
	VectorNormalize( dst );
}

#ifdef _WIN32
#pragma optimize( "", off )
#endif


void RotatePointAroundVector( vec3_t dst, const vec3_t dir, const vec3_t point, float degrees )
{
	float	m[3][3];
	float	im[3][3];
	float	zrot[3][3];
	float	tmpmat[3][3];
	float	rot[3][3];
	int	i;
	vec3_t vr, vup, vf;

	vf[0] = dir[0];
	vf[1] = dir[1];
	vf[2] = dir[2];

	PerpendicularVector( vr, dir );
	CrossProduct( vr, vf, vup );

	m[0][0] = vr[0];
	m[1][0] = vr[1];
	m[2][0] = vr[2];

	m[0][1] = vup[0];
	m[1][1] = vup[1];
	m[2][1] = vup[2];

	m[0][2] = vf[0];
	m[1][2] = vf[1];
	m[2][2] = vf[2];

	memcpy( im, m, sizeof( im ) );

	im[0][1] = m[1][0];
	im[0][2] = m[2][0];
	im[1][0] = m[0][1];
	im[1][2] = m[2][1];
	im[2][0] = m[0][2];
	im[2][1] = m[1][2];

	memset( zrot, 0, sizeof( zrot ) );
	zrot[0][0] = zrot[1][1] = zrot[2][2] = 1.0F;

	zrot[0][0] = cos( DEG2RAD( degrees ) );
	zrot[0][1] = sin( DEG2RAD( degrees ) );
	zrot[1][0] = -sin( DEG2RAD( degrees ) );
	zrot[1][1] = cos( DEG2RAD( degrees ) );

	R_ConcatRotations( m, zrot, tmpmat );
	R_ConcatRotations( tmpmat, im, rot );

	for ( i = 0; i < 3; i++ )
	{
		dst[i] = rot[i][0] * point[0] + rot[i][1] * point[1] + rot[i][2] * point[2];
	}
}

#ifdef _WIN32
#pragma optimize( "", on )
#endif

/*-----------------------------------------------------------------*/

float	anglemod(float a)
{
#if 0
	if (a >= 0)
		a -= 360*(int)(a/360);
	else
		a += 360*( 1 + (int)(-a/360) );
#endif
	a = (360.0/65536) * ((int)(a*(65536/360.0)) & 65535);
	return a;
}

/*
==================
BOPS_Error

Split out like this for ASM to call.
==================
*/
void BOPS_Error (void)
{
	Sys_Error ("BoxOnPlaneSide:  Bad signbits");
}

#if !id386

/*
==================
BoxOnPlaneSide

Returns 1, 2, or 1 + 2
==================
*/
int BoxOnPlaneSide (vec3_t emins, vec3_t emaxs, mplane_t *p)
{
	float	dist1, dist2;
	int		sides;

#if 0	// this is done by the BOX_ON_PLANE_SIDE macro before calling this
		// function
// fast axial cases
	if (p->type < 3)
	{
		if (p->dist <= emins[p->type])
			return 1;
		if (p->dist >= emaxs[p->type])
			return 2;
		return 3;
	}
#endif
	
// general case
	switch (p->signbits)
	{
	case 0:
dist1 = p->normal[0]*emaxs[0] + p->normal[1]*emaxs[1] + p->normal[2]*emaxs[2];
dist2 = p->normal[0]*emins[0] + p->normal[1]*emins[1] + p->normal[2]*emins[2];
		break;
	case 1:
dist1 = p->normal[0]*emins[0] + p->normal[1]*emaxs[1] + p->normal[2]*emaxs[2];
dist2 = p->normal[0]*emaxs[0] + p->normal[1]*emins[1] + p->normal[2]*emins[2];
		break;
	case 2:
dist1 = p->normal[0]*emaxs[0] + p->normal[1]*emins[1] + p->normal[2]*emaxs[2];
dist2 = p->normal[0]*emins[0] + p->normal[1]*emaxs[1] + p->normal[2]*emins[2];
		break;
	case 3:
dist1 = p->normal[0]*emins[0] + p->normal[1]*emins[1] + p->normal[2]*emaxs[2];
dist2 = p->normal[0]*emaxs[0] + p->normal[1]*emaxs[1] + p->normal[2]*emins[2];
		break;
	case 4:
dist1 = p->normal[0]*emaxs[0] + p->normal[1]*emaxs[1] + p->normal[2]*emins[2];
dist2 = p->normal[0]*emins[0] + p->normal[1]*emins[1] + p->normal[2]*emaxs[2];
		break;
	case 5:
dist1 = p->normal[0]*emins[0] + p->normal[1]*emaxs[1] + p->normal[2]*emins[2];
dist2 = p->normal[0]*emaxs[0] + p->normal[1]*emins[1] + p->normal[2]*emaxs[2];
		break;
	case 6:
dist1 = p->normal[0]*emaxs[0] + p->normal[1]*emins[1] + p->normal[2]*emins[2];
dist2 = p->normal[0]*emins[0] + p->normal[1]*emaxs[1] + p->normal[2]*emaxs[2];
		break;
	case 7:
dist1 = p->normal[0]*emins[0] + p->normal[1]*emins[1] + p->normal[2]*emins[2];
dist2 = p->normal[0]*emaxs[0] + p->normal[1]*emaxs[1] + p->normal[2]*emaxs[2];
		break;
	default:
		dist1 = dist2 = 0;		// shut up compiler
		BOPS_Error ();
		break;
	}

#if 0
	int		i;
	vec3_t	corners[2];

	for (i=0 ; i<3 ; i++)
	{
		if (plane->normal[i] < 0)
		{
			corners[0][i] = emins[i];
			corners[1][i] = emaxs[i];
		}
		else
		{
			corners[1][i] = emins[i];
			corners[0][i] = emaxs[i];
		}
	}
	dist = DotProduct (plane->normal, corners[0]) - plane->dist;
	dist2 = DotProduct (plane->normal, corners[1]) - plane->dist;
	sides = 0;
	if (dist1 >= 0)
		sides = 1;
	if (dist2 < 0)
		sides |= 2;

#endif

	sides = 0;
	if (dist1 >= p->dist)
		sides = 1;
	if (dist2 < p->dist)
		sides |= 2;

#ifdef PARANOID
if (sides == 0)
	Sys_Error ("BoxOnPlaneSide: sides==0");
#endif

	return sides;
}

#endif


void AngleVectors (vec3_t angles, vec3_t forward, vec3_t right, vec3_t up)
{
	float		angle;
	float		sr, sp, sy, cr, cp, cy;
	
	angle = angles[YAW] * (M_PI*2 / 360);
	sy = sin(angle);
	cy = cos(angle);
	angle = angles[PITCH] * (M_PI*2 / 360);
	sp = sin(angle);
	cp = cos(angle);
	angle = angles[ROLL] * (M_PI*2 / 360);
	sr = sin(angle);
	cr = cos(angle);

	if (forward)
	{ //rww
		forward[0] = cp*cy;
		forward[1] = cp*sy;
		forward[2] = -sp;
	}
	if (right)
	{ //rww
		right[0] = (-1*sr*sp*cy+-1*cr*-sy);
		right[1] = (-1*sr*sp*sy+-1*cr*cy);
		right[2] = -1*sr*cp;
	}
	if (up)
	{ //rww
		up[0] = (cr*sp*cy+-sr*-sy);
		up[1] = (cr*sp*sy+-sr*cy);
		up[2] = cr*cp;
	}
}

int VectorCompare (vec3_t v1, vec3_t v2)
{
	int		i;
	
	for (i=0 ; i<3 ; i++)
		if (v1[i] != v2[i])
			return 0;
			
	return 1;
}

void VectorMA (vec3_t veca, float scale, vec3_t vecb, vec3_t vecc)
{
	vecc[0] = veca[0] + scale*vecb[0];
	vecc[1] = veca[1] + scale*vecb[1];
	vecc[2] = veca[2] + scale*vecb[2];
}


vec_t _DotProduct (vec3_t v1, vec3_t v2)
{
	return v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2];
}

void _VectorSubtract (vec3_t veca, vec3_t vecb, vec3_t out)
{
	out[0] = veca[0]-vecb[0];
	out[1] = veca[1]-vecb[1];
	out[2] = veca[2]-vecb[2];
}

void _VectorAdd (vec3_t veca, vec3_t vecb, vec3_t out)
{
	out[0] = veca[0]+vecb[0];
	out[1] = veca[1]+vecb[1];
	out[2] = veca[2]+vecb[2];
}

void _VectorCopy (vec3_t in, vec3_t out)
{
	out[0] = in[0];
	out[1] = in[1];
	out[2] = in[2];
}

void CrossProduct (vec3_t v1, vec3_t v2, vec3_t cross)
{
	cross[0] = v1[1]*v2[2] - v1[2]*v2[1];
	cross[1] = v1[2]*v2[0] - v1[0]*v2[2];
	cross[2] = v1[0]*v2[1] - v1[1]*v2[0];
}

double sqrt(double x);

vec_t Length(vec3_t v)
{
	int		i;
	float	length;
	
	length = 0;
	for (i=0 ; i< 3 ; i++)
		length += v[i]*v[i];
	length = sqrt (length);		// FIXME

	return length;
}

float VectorNormalize (vec3_t v)
{
	float	length, ilength;

	length = v[0]*v[0] + v[1]*v[1] + v[2]*v[2];
	length = sqrt (length);		// FIXME

	if (length)
	{
		ilength = 1/length;
		v[0] *= ilength;
		v[1] *= ilength;
		v[2] *= ilength;
	}
		
	return length;

}

void VectorInverse (vec3_t v)
{
	v[0] = -v[0];
	v[1] = -v[1];
	v[2] = -v[2];
}

void VectorScale (vec3_t in, vec_t scale, vec3_t out)
{
	out[0] = in[0]*scale;
	out[1] = in[1]*scale;
	out[2] = in[2]*scale;
}


int Q_log2(int val)
{
	int answer=0;
	while ((val>>=1) != 0)
		answer++;
	return answer;
}

//rww
void PlaneEquation(float *x, float *y, float *z, float *plane)
{
	plane[0] = x[1]*(y[2]-z[2]) + y[1]*(z[2]-x[2]) + z[1]*(x[2]-y[2]);
	plane[1] = x[2]*(y[0]-z[0]) + y[2]*(z[0]-x[0]) + z[2]*(x[0]-y[0]);
	plane[2] = x[0]*(y[1]-z[1]) + y[0]*(z[1]-x[1]) + z[0]*(x[1]-y[1]);
	plane[3] = -( x[0]*( y[1]*z[2] - z[1]*y[2] ) + y[0]*(z[1]*x[2] - x[1]*z[2]) + z[0]*(x[1]*y[2] - y[1]*x[2]) );
}

//rww
int PointRelativeToPlane(float *pos, float *side, float *planeEq)
{
	float s;
	if (!side)
	{
		side = &s;
	}
	*side = planeEq[0]*pos[0] + planeEq[1]*pos[1] + planeEq[2]*pos[2] + planeEq[3];

	if (*side > 0.0f)
	{
		return 1;
	}
	else if (*side < 0.0f)
	{
		return -1;
	}

	return 0;
}

//rww
qboolean IntersectLinePlane(float *start, float *end, float *plane, float *out, float *outDist)
{
	float s1, s2;
	if (PointRelativeToPlane(start, &s1, plane) != PointRelativeToPlane(end, &s2, plane))
	{ //collision on this plane
		float dist;
		float dir[3];

		VectorSubtract(end, start, dir);
		VectorNormalize(dir);

		s2 = plane[0]*dir[0] + plane[1]*dir[1] + plane[2]*dir[2];
		dist = s1/s2;

		VectorMA(start, -dist, dir, out);

		if (outDist)
		{
			*outDist = fabsf(dist);
		}

		return true;
	}

	return false;
}

//rww
//this function was completely pulled out of my ass.
//i think the intersect logic may be lacking, results in more points than necessary.
int ChopBoxToFrustum(viewFrustum_t *frustum, float *mins, float *maxs,
						   float *out, int numOut, float chopFactor)
{
	int ptOut = 0;
	float boundsPoints[8][3];
	float boundsAxis[8][9];
	int i, j, k;
	GetBoundsPoints(mins, maxs, (float *)boundsPoints, (float *)boundsAxis);

	for (i = 0; i < 8; i++)
	{
		float side;
		float curPoint[3];
		curPoint[0] = boundsPoints[i][0];
		curPoint[1] = boundsPoints[i][1];
		curPoint[2] = boundsPoints[i][2];
		for (j = 0; j < 6; j++)
		{
			side = (*frustum)[j][0]*curPoint[0]+
				(*frustum)[j][1]*curPoint[1]+
				(*frustum)[j][2]*curPoint[2]+
				(*frustum)[j][3];
			if (side < 0.0f)
			{ //move the point onto the plane
				float axp[3];
				qboolean gotConfined  = false;
				for (k = 0; k < 3; k++)
				{
					float *ax = &boundsAxis[i][k*3];
					float hit[3];
					float dist;
					VectorMA(boundsPoints[i], maxs[0]-mins[0], ax, axp);
					if (IntersectLinePlane(boundsPoints[i], axp, &(*frustum)[j][0], hit, &dist))
					{
						curPoint[0] = hit[0];
						curPoint[1] = hit[1];
						curPoint[2] = hit[2];

						out[0] = curPoint[0];
						out[1] = curPoint[1];
						out[2] = curPoint[2];
						out += 3;
						ptOut++;
						gotConfined = true;
					}
				}
				if (!gotConfined )
				{
					curPoint[0] -= (*frustum)[j][0]*(side*chopFactor);
					curPoint[1] -= (*frustum)[j][1]*(side*chopFactor);
					curPoint[2] -= (*frustum)[j][2]*(side*chopFactor);
				}
			}
		}

		out[0] = curPoint[0];
		out[1] = curPoint[1];
		out[2] = curPoint[2];
		out += 3;
		ptOut++;
	}

	return ptOut;
}

//rww
int BoxInFrustum(viewFrustum_t *frustum, float *mins, float *maxs)
{
	float boundsPoints[8][3];
	float side;
	int pointsInPlanes = 0;
	int i, j;

	GetBoundsPoints(mins, maxs, (float *)boundsPoints, NULL);

	for (i = 0; i < 6; i++)
	{
		int pointsOutside = 0;
		for (j = 0; j < 8; j++)
		{
			side = (*frustum)[i][0]*boundsPoints[j][0]+
				(*frustum)[i][1]*boundsPoints[j][1]+
				(*frustum)[i][2]*boundsPoints[j][2]+
				(*frustum)[i][3];
			if (side <= 0.0f)
			{
				pointsOutside++;
			}
		}

		if (pointsOutside == 8)
		{ //all points were outside this frustum plane
			return 0;
		}

		if (pointsOutside == 0)
		{ //if no points were outside
			pointsInPlanes++;
		}
	}

	if (pointsInPlanes == 6)
	{ //all points were inside all 6 planes
		return 1;
	}

	//intersecting
	return -1;
}

//rww
static __inline void SetDir(int dir, float *v)
{
	switch (dir)
	{
	case 1:
		v[0] = 1.0f;
		v[1] = 0.0f;
		v[2] = 0.0f;
		break;
	case 2:
		v[0] = 0.0f;
		v[1] = 1.0f;
		v[2] = 0.0f;
		break;
	case 3:
		v[0] = 0.0f;
		v[1] = 0.0f;
		v[2] = 1.0f;
		break;
	case -1:
		v[0] = -1.0f;
		v[1] = 0.0f;
		v[2] = 0.0f;
		break;
	case -2:
		v[0] = 0.0f;
		v[1] = -1.0f;
		v[2] = 0.0f;
		break;
	case -3:
		v[0] = 0.0f;
		v[1] = 0.0f;
		v[2] = -1.0f;
		break;
	default:
		break;
	}
}

//rww
void GetBoundsPoints(float *mins, float *maxs, float *bp, float *baxis)
{
	//RED
	bp[0] = mins[0];
	bp[1] = mins[1];
	bp[2] = mins[2];
	bp += 3;
	if (baxis)
	{
		SetDir(1, baxis);
		SetDir(2, baxis+3);
		SetDir(3, baxis+6);
		baxis += 9;
	}

	//GREEN
	bp[0] = mins[0];
	bp[1] = maxs[1];
	bp[2] = mins[2];
	bp += 3;
	if (baxis)
	{
		SetDir(1, baxis);
		SetDir(-2, baxis+3);
		SetDir(3, baxis+6);
		baxis += 9;
	}

	//BLUE
	bp[0] = maxs[0];
	bp[1] = mins[1];
	bp[2] = mins[2];
	bp += 3;
	if (baxis)
	{
		SetDir(-1, baxis);
		SetDir(2, baxis+3);
		SetDir(3, baxis+6);
		baxis += 9;
	}

	//YELLOW
	bp[0] = maxs[0];
	bp[1] = maxs[1];
	bp[2] = mins[2];
	bp += 3;
	if (baxis)
	{
		SetDir(-1, baxis);
		SetDir(-2, baxis+3);
		SetDir(3, baxis+6);
		baxis += 9;
	}
	//=======================
	//PINK
	bp[0] = mins[0];
	bp[1] = mins[1];
	bp[2] = maxs[2];
	bp += 3;
	if (baxis)
	{
		SetDir(1, baxis);
		SetDir(2, baxis+3);
		SetDir(-3, baxis+6);
		baxis += 9;
	}

	//ORANGE
	bp[0] = mins[0];
	bp[1] = maxs[1];
	bp[2] = maxs[2];
	bp += 3;
	if (baxis)
	{
		SetDir(1, baxis);
		SetDir(-2, baxis+3);
		SetDir(-3, baxis+6);
		baxis += 9;
	}

	//LIGHT BLUE
	bp[0] = maxs[0];
	bp[1] = mins[1];
	bp[2] = maxs[2];
	bp += 3;
	if (baxis)
	{
		SetDir(-1, baxis);
		SetDir(2, baxis+3);
		SetDir(-3, baxis+6);
		baxis += 9;
	}

	//WHITE
	bp[0] = maxs[0];
	bp[1] = maxs[1];
	bp[2] = maxs[2];
	if (baxis)
	{
		SetDir(-1, baxis);
		SetDir(-2, baxis+3);
		SetDir(-3, baxis+6);
		baxis += 9;
	}
}

//rww
qboolean BoxesOverlap(float *mins1, float *maxs1, float *mins2, float *maxs2)
{
	int i = 0;
	while (i < 3)
	{
		if (mins1[i] < mins2[i] &&
			maxs1[i] < mins2[i])
		{
			return false;
		}
		if (mins1[i] > maxs2[i] &&
			maxs1[i] > maxs2[i])
		{
			return false;
		}

		i++;
	}

	return true;
}

//rww
float PlaneNormalize(float *plane)
{
	float length;

	length = plane[0]*plane[0] + plane[1]*plane[1] + plane[2]*plane[2];
	if (length == 0.0f)
	{
		return 0.0f;
	}
	length = sqrtf(length);
	plane[0] = plane[0]/length;
	plane[1] = plane[1]/length;
	plane[2] = plane[2]/length;
	plane[3] = plane[3]/length;

	return length;
}

//rww - taken from some p3 instruction optimization paper
void MatrixInverse4x4(float *mat, float *dst)
{
	float tmp[12]; /* temp array for pairs */
	float src[16]; /* array of transpose source matrix */
	float det; /* determinant */
	int i, j;
	/* transpose matrix */
	for ( i = 0; i < 4; i++)
	{
		src[i] = mat[i*4];
		src[i + 4] = mat[i*4 + 1];
		src[i + 8] = mat[i*4 + 2];
		src[i + 12] = mat[i*4 + 3];
	}
	/* calculate pairs for first 8 elements (cofactors) */
	tmp[0] = src[10] * src[15];
	tmp[1] = src[11] * src[14];
	tmp[2] = src[9] * src[15];
	tmp[3] = src[11] * src[13];
	tmp[4] = src[9] * src[14];
	tmp[5] = src[10] * src[13];
	tmp[6] = src[8] * src[15];
	tmp[7] = src[11] * src[12];
	tmp[8] = src[8] * src[14];
	tmp[9] = src[10] * src[12];
	tmp[10] = src[8] * src[13];
	tmp[11] = src[9] * src[12];
	/* calculate first 8 elements (cofactors) */
	dst[0] = tmp[0]*src[5] + tmp[3]*src[6] + tmp[4]*src[7];
	dst[0] -= tmp[1]*src[5] + tmp[2]*src[6] + tmp[5]*src[7];
	dst[1] = tmp[1]*src[4] + tmp[6]*src[6] + tmp[9]*src[7];
	dst[1] -= tmp[0]*src[4] + tmp[7]*src[6] + tmp[8]*src[7];
	dst[2] = tmp[2]*src[4] + tmp[7]*src[5] + tmp[10]*src[7];
	dst[2] -= tmp[3]*src[4] + tmp[6]*src[5] + tmp[11]*src[7];
	dst[3] = tmp[5]*src[4] + tmp[8]*src[5] + tmp[11]*src[6];
	dst[3] -= tmp[4]*src[4] + tmp[9]*src[5] + tmp[10]*src[6];
	dst[4] = tmp[1]*src[1] + tmp[2]*src[2] + tmp[5]*src[3];
	dst[4] -= tmp[0]*src[1] + tmp[3]*src[2] + tmp[4]*src[3];
	dst[5] = tmp[0]*src[0] + tmp[7]*src[2] + tmp[8]*src[3];
	dst[5] -= tmp[1]*src[0] + tmp[6]*src[2] + tmp[9]*src[3];
	dst[6] = tmp[3]*src[0] + tmp[6]*src[1] + tmp[11]*src[3];
	dst[6] -= tmp[2]*src[0] + tmp[7]*src[1] + tmp[10]*src[3];
	dst[7] = tmp[4]*src[0] + tmp[9]*src[1] + tmp[10]*src[2];
	dst[7] -= tmp[5]*src[0] + tmp[8]*src[1] + tmp[11]*src[2];
	/* calculate pairs for second 8 elements (cofactors) */
	tmp[0] = src[2]*src[7];
	tmp[1] = src[3]*src[6];
	tmp[2] = src[1]*src[7];
	tmp[3] = src[3]*src[5];
	tmp[4] = src[1]*src[6];
	tmp[5] = src[2]*src[5];
	tmp[6] = src[0]*src[7];
	tmp[7] = src[3]*src[4];
	tmp[8] = src[0]*src[6];
	tmp[9] = src[2]*src[4];
	tmp[10] = src[0]*src[5];
	tmp[11] = src[1]*src[4];
	/* calculate second 8 elements (cofactors) */
	dst[8] = tmp[0]*src[13] + tmp[3]*src[14] + tmp[4]*src[15];
	dst[8] -= tmp[1]*src[13] + tmp[2]*src[14] + tmp[5]*src[15];
	dst[9] = tmp[1]*src[12] + tmp[6]*src[14] + tmp[9]*src[15];
	dst[9] -= tmp[0]*src[12] + tmp[7]*src[14] + tmp[8]*src[15];
	dst[10] = tmp[2]*src[12] + tmp[7]*src[13] + tmp[10]*src[15];
	dst[10]-= tmp[3]*src[12] + tmp[6]*src[13] + tmp[11]*src[15];
	dst[11] = tmp[5]*src[12] + tmp[8]*src[13] + tmp[11]*src[14];
	dst[11]-= tmp[4]*src[12] + tmp[9]*src[13] + tmp[10]*src[14];
	dst[12] = tmp[2]*src[10] + tmp[5]*src[11] + tmp[1]*src[9];
	dst[12]-= tmp[4]*src[11] + tmp[0]*src[9] + tmp[3]*src[10];
	dst[13] = tmp[8]*src[11] + tmp[0]*src[8] + tmp[7]*src[10];
	dst[13]-= tmp[6]*src[10] + tmp[9]*src[11] + tmp[1]*src[8];
	dst[14] = tmp[6]*src[9] + tmp[11]*src[11] + tmp[3]*src[8];
	dst[14]-= tmp[10]*src[11] + tmp[2]*src[8] + tmp[7]*src[9];
	dst[15] = tmp[10]*src[10] + tmp[4]*src[8] + tmp[9]*src[9];
	dst[15]-= tmp[8]*src[9] + tmp[11]*src[10] + tmp[5]*src[8];
	/* calculate determinant */
	det=src[0]*dst[0]+src[1]*dst[1]+src[2]*dst[2]+src[3]*dst[3];
	/* calculate matrix inverse */
	if (det != 0.0f)
	{
		det = 1/det;
	}
	for ( j = 0; j < 16; j++)
		dst[j] *= det;
}

//rww
void TranslateMatrix4x4(float *mat, float *v)
{
	mat[12+0] += v[0]*mat[0];
	mat[12+1] += v[0]*mat[1];
	mat[12+2] += v[0]*mat[2];

	mat[12+0] += v[1]*mat[4+0];
	mat[12+1] += v[1]*mat[4+1];
	mat[12+2] += v[1]*mat[4+2];

	mat[12+0] += v[2]*mat[8+0];
	mat[12+1] += v[2]*mat[8+1];
	mat[12+2] += v[2]*mat[8+2];
}

//rww
void MatrixMultiply4x4(float *in, float *in2, float *out)
{
	out[0] = (in[0]*in2[0]) + (in[4+0]*in2[1]) + (in[8+0]*in2[2]) + (in[12+0]*in2[3]);
	out[4+0] = (in[0]*in2[4+0]) + (in[4+0]*in2[4+1]) + (in[8+0]*in2[4+2]) + (in[12+0]*in2[4+3]);
	out[8+0] = (in[0]*in2[8+0]) + (in[4+0]*in2[8+1]) + (in[8+0]*in2[8+2]) + (in[12+0]*in2[8+3]);
	out[12+0] = (in[0]*in2[12+0]) + (in[4+0]*in2[12+1]) + (in[8+0]*in2[12+2]) + (in[12+0]*in2[12+3]);

	out[1] = (in[1]*in2[0]) + (in[4+1]*in2[1]) + (in[8+1]*in2[2]) + (in[12+1]*in2[3]);
	out[4+1] = (in[1]*in2[4+0]) + (in[4+1]*in2[4+1]) + (in[8+1]*in2[4+2]) + (in[12+1]*in2[4+3]);
	out[8+1] = (in[1]*in2[8+0]) + (in[4+1]*in2[8+1]) + (in[8+1]*in2[8+2]) + (in[12+1]*in2[8+3]);
	out[12+1] = (in[1]*in2[12+0]) + (in[4+1]*in2[12+1]) + (in[8+1]*in2[12+2]) + (in[12+1]*in2[12+3]);

	out[2] = (in[2]*in2[0]) + (in[4+2]*in2[1]) + (in[8+2]*in2[2]) + (in[12+2]*in2[3]);
	out[4+2] = (in[2]*in2[4+0]) + (in[4+2]*in2[4+1]) + (in[8+2]*in2[4+2]) + (in[12+2]*in2[4+3]);
	out[8+2] = (in[2]*in2[8+0]) + (in[4+2]*in2[8+1]) + (in[8+2]*in2[8+2]) + (in[12+2]*in2[8+3]);
	out[12+2] = (in[2]*in2[12+0]) + (in[4+2]*in2[12+1]) + (in[8+2]*in2[12+2]) + (in[12+2]*in2[12+3]);

	out[3] = (in[3]*in2[0]) + (in[4+3]*in2[1]) + (in[8+3]*in2[2]) + (in[12+3]*in2[3]);
	out[4+3] = (in[3]*in2[4+0]) + (in[4+3]*in2[4+1]) + (in[8+3]*in2[4+2]) + (in[12+3]*in2[4+3]);
	out[8+3] = (in[3]*in2[8+0]) + (in[4+3]*in2[8+1]) + (in[8+3]*in2[8+2]) + (in[12+3]*in2[8+3]);
	out[12+3] = (in[3]*in2[12+0]) + (in[4+3]*in2[12+1]) + (in[8+3]*in2[12+2]) + (in[12+3]*in2[12+3]);
}

//rww
void TransformPointByMatrix4x4(float *matrix, float *in, float *out)
{
	out[0] = matrix[0]*in[0] + matrix[4+0]*in[1] + matrix[8+0]*in[2] + matrix[12+0];
	out[1] = matrix[1]*in[0] + matrix[4+1]*in[1] + matrix[8+1]*in[2] + matrix[12+1];
	out[2] = matrix[2]*in[0] + matrix[4+2]*in[1] + matrix[8+2]*in[2] + matrix[12+2];
}

//rww
void TransformV4ByMatrix4x4(float *matrix, float *in, float *out)
{
	out[0] = matrix[0]*in[0] + matrix[4+0]*in[1] + matrix[8+0]*in[2] + matrix[12+0]*in[3];
	out[1] = matrix[1]*in[0] + matrix[4+1]*in[1] + matrix[8+1]*in[2] + matrix[12+1]*in[3];
	out[2] = matrix[2]*in[0] + matrix[4+2]*in[1] + matrix[8+2]*in[2] + matrix[12+2]*in[3];
	out[3] = matrix[3]*in[0] + matrix[4+3]*in[1] + matrix[8+3]*in[2] + matrix[12+3]*in[3];
}

/*
================
R_ConcatRotations
================
*/
void R_ConcatRotations (float in1[3][3], float in2[3][3], float out[3][3])
{
	out[0][0] = in1[0][0] * in2[0][0] + in1[0][1] * in2[1][0] +
				in1[0][2] * in2[2][0];
	out[0][1] = in1[0][0] * in2[0][1] + in1[0][1] * in2[1][1] +
				in1[0][2] * in2[2][1];
	out[0][2] = in1[0][0] * in2[0][2] + in1[0][1] * in2[1][2] +
				in1[0][2] * in2[2][2];
	out[1][0] = in1[1][0] * in2[0][0] + in1[1][1] * in2[1][0] +
				in1[1][2] * in2[2][0];
	out[1][1] = in1[1][0] * in2[0][1] + in1[1][1] * in2[1][1] +
				in1[1][2] * in2[2][1];
	out[1][2] = in1[1][0] * in2[0][2] + in1[1][1] * in2[1][2] +
				in1[1][2] * in2[2][2];
	out[2][0] = in1[2][0] * in2[0][0] + in1[2][1] * in2[1][0] +
				in1[2][2] * in2[2][0];
	out[2][1] = in1[2][0] * in2[0][1] + in1[2][1] * in2[1][1] +
				in1[2][2] * in2[2][1];
	out[2][2] = in1[2][0] * in2[0][2] + in1[2][1] * in2[1][2] +
				in1[2][2] * in2[2][2];
}


/*
================
R_ConcatTransforms
================
*/
void R_ConcatTransforms (float in1[3][4], float in2[3][4], float out[3][4])
{
	out[0][0] = in1[0][0] * in2[0][0] + in1[0][1] * in2[1][0] +
				in1[0][2] * in2[2][0];
	out[0][1] = in1[0][0] * in2[0][1] + in1[0][1] * in2[1][1] +
				in1[0][2] * in2[2][1];
	out[0][2] = in1[0][0] * in2[0][2] + in1[0][1] * in2[1][2] +
				in1[0][2] * in2[2][2];
	out[0][3] = in1[0][0] * in2[0][3] + in1[0][1] * in2[1][3] +
				in1[0][2] * in2[2][3] + in1[0][3];
	out[1][0] = in1[1][0] * in2[0][0] + in1[1][1] * in2[1][0] +
				in1[1][2] * in2[2][0];
	out[1][1] = in1[1][0] * in2[0][1] + in1[1][1] * in2[1][1] +
				in1[1][2] * in2[2][1];
	out[1][2] = in1[1][0] * in2[0][2] + in1[1][1] * in2[1][2] +
				in1[1][2] * in2[2][2];
	out[1][3] = in1[1][0] * in2[0][3] + in1[1][1] * in2[1][3] +
				in1[1][2] * in2[2][3] + in1[1][3];
	out[2][0] = in1[2][0] * in2[0][0] + in1[2][1] * in2[1][0] +
				in1[2][2] * in2[2][0];
	out[2][1] = in1[2][0] * in2[0][1] + in1[2][1] * in2[1][1] +
				in1[2][2] * in2[2][1];
	out[2][2] = in1[2][0] * in2[0][2] + in1[2][1] * in2[1][2] +
				in1[2][2] * in2[2][2];
	out[2][3] = in1[2][0] * in2[0][3] + in1[2][1] * in2[1][3] +
				in1[2][2] * in2[2][3] + in1[2][3];
}


/*
===================
FloorDivMod

Returns mathematically correct (floor-based) quotient and remainder for
numer and denom, both of which should contain no fractional part. The
quotient must fit in 32 bits.
====================
*/

void FloorDivMod (double numer, double denom, int *quotient,
		int *rem)
{
	int		q, r;
	double	x;

#ifndef PARANOID
	if (denom <= 0.0)
		Sys_Error ("FloorDivMod: bad denominator %d\n", denom);

//	if ((floor(numer) != numer) || (floor(denom) != denom))
//		Sys_Error ("FloorDivMod: non-integer numer or denom %f %f\n",
//				numer, denom);
#endif

	if (numer >= 0.0)
	{

		x = floor(numer / denom);
		q = (int)x;
		r = (int)floor(numer - (x * denom));
	}
	else
	{
	//
	// perform operations with positive values, and fix mod to make floor-based
	//
		x = floor(-numer / denom);
		q = -(int)x;
		r = (int)floor(-numer - (x * denom));
		if (r != 0)
		{
			q--;
			r = (int)denom - r;
		}
	}

	*quotient = q;
	*rem = r;
}


/*
===================
GreatestCommonDivisor
====================
*/
int GreatestCommonDivisor (int i1, int i2)
{
	if (i1 > i2)
	{
		if (i2 == 0)
			return (i1);
		return GreatestCommonDivisor (i2, i1 % i2);
	}
	else
	{
		if (i1 == 0)
			return (i2);
		return GreatestCommonDivisor (i1, i2 % i1);
	}
}


#if !id386

// TODO: move to nonintel.c

/*
===================
Invert24To16

Inverts an 8.24 value to a 16.16 value
====================
*/

fixed16_t Invert24To16(fixed16_t val)
{
	if (val < 256)
		return (0xFFFFFFFF);

	return (fixed16_t)
			(((double)0x10000 * (double)0x1000000 / (double)val) + 0.5);
}

#endif
